//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Detector/IDetector.cpp
//! @brief     Implements common detector interface.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Detector/IDetector.h"
#include "Base/Utils/Assert.h"
#include "Device/ProDetector/DetectorMask.h"
#include "Device/Resolution/ConvolutionDetectorResolution.h"
#include "Resample/Element/DiffuseElement.h"

IDetector::IDetector()
{
    registerChild(&m_detection_properties);
}

IDetector::IDetector(const IDetector& other)
    : ICloneable()
    , INode()
    , m_explicitROI(other.m_explicitROI)
    , m_axes(other.m_axes)
    , m_detection_properties(other.m_detection_properties)
{
    if (other.m_detector_resolution)
        setDetectorResolution(*other.m_detector_resolution);
    setName(other.getName());
    registerChild(&m_detection_properties);
}

IDetector::~IDetector() = default;

void IDetector::addAxis(const IAxis& axis)
{
    m_axes.push_back(axis.clone());
}

size_t IDetector::dimension() const
{
    return m_axes.size();
}

void IDetector::clear()
{
    m_axes.clear();
}

const IAxis& IDetector::axis(size_t index) const
{
    ASSERT(index < dimension());
    return *m_axes[index];
}

size_t IDetector::axisBinIndex(size_t index, size_t selected_axis) const
{
    const size_t dim = dimension();
    size_t remainder(index);
    size_t i_axis = dim;
    for (size_t i = 0; i < dim; ++i) {
        --i_axis;
        if (selected_axis == i_axis)
            return remainder % m_axes[i_axis]->size();
        remainder /= m_axes[i_axis]->size();
    }
    ASSERT(0);
}

std::unique_ptr<IAxis> IDetector::createAxis(size_t index, size_t n_bins, double min,
                                             double max) const
{
    if (max <= min)
        throw std::runtime_error("IDetector::createAxis() -> Error! max <= min");
    if (n_bins == 0)
        throw std::runtime_error("IDetector::createAxis() -> Error! Number n_bins can't be zero.");
    return std::make_unique<FixedBinAxis>(axisName(index), n_bins, min, max);
}

size_t IDetector::sizeOfExplicitRegionOfInterest() const
{
    if (m_explicitROI.size() != m_axes.size())
        return 0;

    size_t s = 1;
    for (const auto& roiOfAxis : m_explicitROI)
        s *= roiOfAxis.roiSize;

    return s;
}

std::pair<double, double> IDetector::boundsOfExplicitRegionOfInterest(size_t iAxis) const
{
    ASSERT(iAxis < dimension());
    if (iAxis < m_explicitROI.size())
        return {m_explicitROI[iAxis].lower, m_explicitROI[iAxis].upper};
    return {0., 0.};
}

size_t IDetector::totalSize() const
{
    const size_t dim = dimension();
    if (dim == 0)
        return 0;
    size_t result = 1;
    for (size_t i_axis = 0; i_axis < dim; ++i_axis)
        result *= m_axes[i_axis]->size();
    return result;
}

size_t IDetector::sizeOfRegionOfInterest() const
{
    const auto explicitSize = sizeOfExplicitRegionOfInterest();
    return (explicitSize != 0) ? explicitSize : totalSize();
}

bool IDetector::hasExplicitRegionOfInterest() const
{
    return !m_axes.empty() && (m_explicitROI.size() == m_axes.size());
}

CloneableVector<IAxis> IDetector::axesClippedToRegionOfInterest() const
{
    CloneableVector<IAxis> result;
    for (size_t iAxis = 0; iAxis < m_axes.size(); ++iAxis) {
        auto axis = m_axes[iAxis]->clone();
        axis->clip(regionOfInterestBounds(iAxis));
        result.push_back(axis);
    }
    return result;
}

size_t IDetector::regionOfInterestIndexToDetectorIndex(const size_t regionOfInterestIndex) const
{
    if (!hasExplicitRegionOfInterest())
        return regionOfInterestIndex;

    throw std::runtime_error("IDetector::regionOfInterestIndexToDetectorIndex() -> Error. Not "
                             "implemented by derived class.");
}

size_t IDetector::detectorIndexToRegionOfInterestIndex(const size_t detectorIndex) const
{
    if (!hasExplicitRegionOfInterest())
        return detectorIndex;

    throw std::runtime_error("IDetector::detectorIndexToRegionOfInterestIndex() -> Error. Not "
                             "implemented by derived class.");
}

void IDetector::setAnalyzerProperties(const kvector_t direction, double efficiency,
                                      double total_transmission)
{
    m_detection_properties.setAnalyzerProperties(direction, efficiency, total_transmission);
}

void IDetector::setDetectorResolution(const IDetectorResolution& p_detector_resolution)
{
    m_detector_resolution.reset(p_detector_resolution.clone());
    registerChild(m_detector_resolution.get());
}

// TODO: pass dimension-independent argument to this function
void IDetector::setResolutionFunction(const IResolutionFunction2D& resFunc)
{
    ConvolutionDetectorResolution convFunc(resFunc);
    setDetectorResolution(convFunc);
}

void IDetector::resetRegionOfInterest()
{
    m_explicitROI.clear();
}

void IDetector::applyDetectorResolution(OutputData<double>* p_intensity_map) const
{
    ASSERT(p_intensity_map);

    if (m_detector_resolution) {
        m_detector_resolution->applyDetectorResolution(p_intensity_map);
        if (detectorMask() && detectorMask()->hasMasks()) {
            // sets amplitude in masked areas to zero
            std::unique_ptr<OutputData<double>> buff(new OutputData<double>());
            buff->copyShapeFrom(*p_intensity_map);

            iterateOverNonMaskedPoints([&](const_iterator it) {
                (*buff)[it.roiIndex()] = (*p_intensity_map)[it.roiIndex()];
            });
            p_intensity_map->setRawDataVector(buff->getRawDataVector());
        }
    }
}

const IDetectorResolution* IDetector::detectorResolution() const
{
    return m_detector_resolution.get();
}

OutputData<double>* IDetector::createDetectorIntensity(
    const std::vector<std::unique_ptr<DiffuseElement>>& elements) const
{
    std::unique_ptr<OutputData<double>> detectorMap(createDetectorMap());
    ASSERT(detectorMap);
    setDataToDetectorMap(*detectorMap, elements);
    if (m_detector_resolution)
        applyDetectorResolution(detectorMap.get());

    return detectorMap.release();
}

std::unique_ptr<OutputData<double>> IDetector::createDetectorMap() const
{
    const size_t dim = dimension();
    ASSERT(dim != 0);

    std::unique_ptr<OutputData<double>> result(new OutputData<double>);
    for (size_t iAxis = 0; iAxis < dim; ++iAxis) {
        std::unique_ptr<IAxis> tempAxis(axis(iAxis).clone());
        tempAxis->clip(regionOfInterestBounds(iAxis));
        result->addAxis(*tempAxis);
    }

    return result;
}

void IDetector::setDataToDetectorMap(
    OutputData<double>& detectorMap,
    const std::vector<std::unique_ptr<DiffuseElement>>& elements) const
{
    if (elements.empty())
        return;

    size_t elementIndex = 0;
    for (auto it = beginNonMaskedPoints(); it != endNonMaskedPoints(); ++it)
        detectorMap[it.roiIndex()] = elements[elementIndex++]->intensity();
}

size_t IDetector::numberOfElements() const
{
    size_t result(0);
    iterateOverNonMaskedPoints([&result](const_iterator) { ++result; });
    return result;
}

std::pair<double, double> IDetector::regionOfInterestBounds(size_t iAxis) const
{
    ASSERT(iAxis < m_axes.size());

    const auto explicitBounds = boundsOfExplicitRegionOfInterest(iAxis);
    if (explicitBounds.first != 0 || explicitBounds.second != 0)
        return explicitBounds;

    return m_axes[iAxis]->bounds();
}

std::vector<const INode*> IDetector::getChildren() const
{
    return std::vector<const INode*>() << &m_detection_properties << m_detector_resolution;
}

void IDetector::iterateOverRegionOfInterest(std::function<void(const_iterator)> func) const
{
    if (this->dimension() == 0)
        return;

    for (auto it = beginRegionOfInterestPoints(); it != endRegionOfInterestPoints(); ++it)
        func(it);
}

void IDetector::iterateOverNonMaskedPoints(std::function<void(const_iterator)> func) const
{
    if (this->dimension() == 0)
        return;

    for (auto it = beginNonMaskedPoints(); it != endNonMaskedPoints(); ++it)
        func(it);
}

SimulationAreaIterator IDetector::beginNonMaskedPoints() const
{
    return SimulationAreaIterator::createBegin(this, SimulationAreaIterator::notMasked);
}

SimulationAreaIterator IDetector::endNonMaskedPoints() const
{
    return SimulationAreaIterator::createEnd(this, SimulationAreaIterator::notMasked);
}

SimulationAreaIterator IDetector::beginRegionOfInterestPoints() const
{
    return SimulationAreaIterator::createBegin(this, SimulationAreaIterator::regionOfInterest);
}

SimulationAreaIterator IDetector::endRegionOfInterestPoints() const
{
    return SimulationAreaIterator::createEnd(this, SimulationAreaIterator::regionOfInterest);
}

/* -- IDetector::RoiOfAxis ----------------------------------------------------------- */

IDetector::RoiOfAxis::RoiOfAxis(const IAxis& axis, double _lower, double _upper)
{
    lower = _lower;
    upper = _upper;

    lowerIndex = axis.findClosestIndex(lower);
    upperIndex = axis.findClosestIndex(upper);

    detectorSize = axis.size();
    roiSize = upperIndex - lowerIndex + 1;
}
