//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Instrument/Instrument.cpp
//! @brief     Implements class Instrument.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Instrument/Instrument.h"
#include "Base/Element/PolarizerPair.h"
#include "Base/Pixel/RectangularPixel.h"
#include "Device/Beam/Beam.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Detector/RectangularDetector.h"
#include "Device/Detector/SphericalDetector.h"
#include "Device/Histo/Histogram2D.h"
#include "Device/Resolution/IResolutionFunction2D.h"

Instrument::Instrument(const Beam& beam, const IDetector& detector)
    : m_beam(beam.clone()), m_detector(detector.clone())
{
    setName("Instrument");
    registerChild(m_detector.get());
    registerChild(m_beam.get());
    initDetector();
}

Instrument::Instrument() : Instrument(Beam::horizontalBeam(), SphericalDetector()) {}

Instrument::Instrument(const Instrument& other) : Instrument(other.beam(), other.detector()) {}

Instrument::~Instrument() = default;

Instrument& Instrument::operator=(const Instrument& other)
{
    if (this != &other) {
        if (other.m_beam)
            setBeam(*other.m_beam);
        else
            m_beam.release();
        if (other.m_detector)
            setDetector(*other.m_detector);
        else
            m_detector.release();
    }
    return *this;
}

void Instrument::setDetector(const IDetector& detector)
{
    m_detector.reset(detector.clone());
    registerChild(m_detector.get());
    initDetector();
}

void Instrument::setBeam(const Beam& beam)
{
    m_beam.reset(beam.clone());
    registerChild(m_beam.get());
    if (m_detector)
        initDetector();
}

void Instrument::initDetector()
{
    if (!m_detector)
        throw std::runtime_error(
            "Instrument::initDetector() -> Error. Detector is not initialized.");
    m_detector->setDetectorNormal(beam().direction().zReflected());
}

std::vector<const INode*> Instrument::getChildren() const
{
    std::vector<const INode*> result;
    result.push_back(m_beam.get());
    if (m_detector)
        result.push_back(m_detector.get());
    return result;
}

PolarizerPair Instrument::polarizerPair() const
{
    return {m_beam->getPolarization(), m_detector->detectionProperties().analyzerOperator()};
}

void Instrument::setBeamParameters(double wavelength, double alpha_i, double phi_i)
{
    m_beam->setWavelength(wavelength);
    m_beam->setDirection({alpha_i, phi_i});
    if (m_detector)
        initDetector();
}

const IDetector* Instrument::getDetector() const
{
    return m_detector.get();
}

IDetector* Instrument::getDetector()
{
    return m_detector.get();
}

const IDetector& Instrument::detector() const
{
    ASSERT(m_detector);
    return *m_detector;
}

IDetector& Instrument::detector()
{
    ASSERT(m_detector);
    return *m_detector;
}

CoordSystem2D* Instrument::createScatteringCoords() const
{
    const IDetector* const detector = getDetector();

    // TODO: use virtual fcts in detector classes instead of dynamic casts here
    if (const auto* const det = dynamic_cast<const SphericalDetector*>(detector))
        return new SphericalCoords(detector->axesClippedToRegionOfInterest(), beam().direction(),
                                   beam().wavelength());

    if (const auto* const det = dynamic_cast<const RectangularDetector*>(detector)) {
        std::unique_ptr<RectangularPixel> roiPixel(det->regionOfInterestPixel());
        return new ImageCoords(detector->axesClippedToRegionOfInterest(), *roiPixel,
                               beam().direction(), beam().wavelength());
    }
    ASSERT(0);
}
